/*
  USART_1WIRE -- Asynchronous 1-wire USART demonstration.
  This uses USART0 (TX on pin D2) and USART1 (TX on pin D1)
  and sends data between the two. Data sent out USART0 is received in USART1 which
  then sends the data back to USART0. Half duplex operation.
  Copyright (C) 2020  Thomas Almy

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Contact the author at tom@almy.us */
#include <Pins.h>

#define BAUDRATE 9600 // Set the baud rate
#define BLOCKSIZE 5   // Number of bytes to send. May not be larger
// than BUFSIZE-1 otherwise program will lock up
#define BUFSIZE 8 // Must be a power of 2!
#define BUFMASK (BUFSIZE-1)
// USART 0 buffers
char putBuffer0[BUFSIZE]; // data to transmit
char getBuffer0[BUFSIZE]; // data to receive
// Indices into buffers
volatile uint8_t getBuffer0InIdx; // data into receive buffer
uint8_t getBuffer0OutIdx; // data out of receive buffer
uint8_t putBuffer0InIdx; // data into transmit buffer
volatile uint8_t putBuffer0OutIdx; // data out of transmit buffer
// USART 1 buffers
char putBuffer1[BUFSIZE]; // data to transmit
char getBuffer1[BUFSIZE]; // data to receive
// Indices into buffers
volatile uint8_t getBuffer1InIdx; // data into receive buffer
uint8_t getBuffer1OutIdx; // data out of receive buffer
uint8_t putBuffer1InIdx; // data into transmit buffer
volatile uint8_t putBuffer1OutIdx; // data out of transmit buffer


void setup() {
  Serial.begin(19200);
  delay(800);

  // TX pins are controlled by USARTs, but we need pullups
  PORTA.PIN0CTRL |= 1 << PORT_PULLUPEN_bp; // Enable pullup D2
  PORTB.PIN4CTRL |= 1 << PORT_PULLUPEN_bp; // Enable pullup D1

  // Configure the port mux
  PORTMUX.USARTROUTEA = PORTMUX_USART1_ALT1_gc | PORTMUX_USART2_ALT1_gc | PORTMUX_USART3_ALT1_gc;

  cli(); // Disable interrupts
  // Configure USART0
  USART0.CTRLC = USART_CHSIZE_8BIT_gc; // 8 bits, no parity 1 stop bit
  USART0.BAUD = (4 * F_CPU) / BAUDRATE;
  USART0.CTRLB = 0xc8; // RXEN TXEN ODME
  USART0.CTRLA = 0x88; // RXCIE, LBME. DREIE turned on when there is something to send

  // Configure USART1
  USART1.CTRLC = USART_CHSIZE_8BIT_gc; // 8 bits, no parity 1 stop bit
  USART1.BAUD = (4 * F_CPU) / BAUDRATE;
  USART1.CTRLB = 0xc8; // RXEN  TXEN ODME
  USART1.CTRLA = 0x88; // RXCIE, LBME. DREIE turned on when there is something to send
  sei(); // reenable interrupts
}

void loop() {
  uint8_t i;
  // Send bytes out USART0
  for (i = 0; i < BLOCKSIZE; i++) {
    putUSART0(i);
  }
  // Receive local echo bytes
  for (i = 0; i < BLOCKSIZE; i++) {
    uint8_t val = getUSART0();
    if (val != i) {
      Serial.print("USART0 Expected Echo ");
      Serial.print(i);
      Serial.print(" received ");
      Serial.println(val);
    }
  }
  // Receive bytes in USART1
  for (i = 0; i < BLOCKSIZE; i++) {
    uint8_t val = getUSART1();
    if (val != i) {
      Serial.print("Expected USART1 ");
      Serial.print(i);
      Serial.print(" received ");
      Serial.println(val);
    }
  }
  delay(1);
  // Now go the other direction
  // Send bytes out USART1
  for (i = 0; i < BLOCKSIZE; i++) {
    putUSART1(i + 64);
  }
  // Receive local echo bytes
  for (i = 0; i < BLOCKSIZE; i++) {
    uint8_t val = getUSART1();
    if (val != i + 64) {
      Serial.print("USART1 Expected Echo ");
      Serial.print(i + 64);
      Serial.print(" received ");
      Serial.println(val);
    }
  }
  // Receive bytes in USART0
  for (i = 0; i < BLOCKSIZE; i++) {
    uint8_t val = getUSART0();
    if (val != i + 64) {
      Serial.print("Expected USART0 ");
      Serial.print(i + 64);
      Serial.print(" received ");
      Serial.println(val);
    }
  }
  delay(1);
}

// USART0 ISRs and functions

ISR(USART0_RXC_vect) {
  if (((getBuffer0InIdx + 1)&BUFMASK) != getBuffer0OutIdx) { // there is room
    getBuffer0[getBuffer0InIdx] = USART0_RXDATAL; // put character in buffer
    getBuffer0InIdx = (getBuffer0InIdx + 1) & BUFMASK; // increment index
  } else {
    // If no room, read the register to toss the character
    (void)USART0_RXDATAL;
  }
}

inline bool isEmpty0() { // receive buffer is empty
  return (getBuffer0InIdx == getBuffer0OutIdx);
}

char getUSART0(void) {
  char retval;
  // block until a character is available
  while (isEmpty0()) {};
  // get character out of buffer
  retval = getBuffer0[getBuffer0OutIdx];
  // increment index
  getBuffer0OutIdx = (getBuffer0OutIdx + 1) & BUFMASK;
  return retval;
}

ISR(USART0_DRE_vect) {
  // There must be a character to send
  USART0.TXDATAL = putBuffer0[putBuffer0OutIdx];
  // Increment the index
  putBuffer0OutIdx = (putBuffer0OutIdx + 1)&BUFMASK;
  // If it is the last character to send, disable the interrupt
  if (putBuffer0OutIdx == putBuffer0InIdx) {
    USART0.CTRLA &= ~USART_DREIE_bm; // turn of DREIE bit
  }
}

inline bool isFull0() { // transmit buffer is full
  return (((putBuffer0InIdx + 1) & BUFMASK) == putBuffer0OutIdx);
}

void putUSART0(char c) {
  while (isFull0()) { // block if buffer is full
    if ((SREG & (1 << SREG_I)) == 0) { // interrupts are disabled so must poll
      if ((USART0.STATUS & USART_DREIF_bm) != 0) { // There is space in buffer
        USART0.TXDATAL = putBuffer0[putBuffer0OutIdx];
        putBuffer0OutIdx = (putBuffer0OutIdx + 1)&BUFMASK;
      }
    }
  }
  putBuffer0[putBuffer0InIdx] = c;
  putBuffer0InIdx = (putBuffer0InIdx + 1) & BUFMASK;
  USART0.CTRLA |= USART_DREIE_bm; // make sure interrupt is enabled
}

// USART1 ISRs and functions

ISR(USART1_RXC_vect) {
  if (((getBuffer1InIdx + 1)&BUFMASK) != getBuffer1OutIdx) { // there is room
    getBuffer1[getBuffer1InIdx] = USART1_RXDATAL; // put character in buffer
    getBuffer1InIdx = (getBuffer1InIdx + 1) & BUFMASK; // increment index
  } else {
    // If no room, read the register to toss the character
    (void)USART1_RXDATAL;
  }
}

inline bool isEmpty1() { // receive buffer is empty
  return (getBuffer1InIdx == getBuffer1OutIdx);
}

char getUSART1(void) {
  char retval;
  // block until a character is available
  while (isEmpty1()) {};
  // get character out of buffer
  retval = getBuffer1[getBuffer1OutIdx];
  // increment index
  getBuffer1OutIdx = (getBuffer1OutIdx + 1) & BUFMASK;
  return retval;
}

ISR(USART1_DRE_vect) {
  // There must be a character to send
  USART1.TXDATAL = putBuffer1[putBuffer1OutIdx];
  // Increment the index
  putBuffer1OutIdx = (putBuffer1OutIdx + 1)&BUFMASK;
  // If it is the last character to send, disable the interrupt
  if (putBuffer1OutIdx == putBuffer1InIdx) {
    USART1.CTRLA &= ~USART_DREIE_bm; // turn of DREIE bit
  }
}

inline bool isFull1() { // transmit buffer is full
  return (((putBuffer1InIdx + 1) & BUFMASK) == putBuffer1OutIdx);
}

void putUSART1(char c) {
  while (isFull1()) { // block if buffer is full
    if ((SREG & (1 << SREG_I)) == 1) { // interrupts are disabled so must poll
      if ((USART1.STATUS & USART_DREIF_bm) != 1) { // There is space in buffer
        USART1.TXDATAL = putBuffer1[putBuffer1OutIdx];
        putBuffer1OutIdx = (putBuffer1OutIdx + 1)&BUFMASK;
      }
    }
  }
  putBuffer1[putBuffer1InIdx] = c;
  putBuffer1InIdx = (putBuffer1InIdx + 1) & BUFMASK;
  USART1.CTRLA |= USART_DREIE_bm; // make sure interrupt is enabled
}
